@@ -1,7 +1,8 @@
 /*
- * afc.c 
+ * afc.c
  * Contains functions for the built-in AFC client.
- * 
+ *
+ * Copyright (c) 2014 Martin Szulecki All Rights Reserved.
  * Copyright (c) 2009-2014 Nikias Bassen. All Rights Reserved.
  * Copyright (c) 2008 Zach C. All Rights Reserved.
  *
@@ -9,15 +10,15 @@
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2.1 of the License, or (at your option) any later version.
- * 
+ *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA 
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
 #ifdef HAVE_CONFIG_H
@@ -35,7 +36,7 @@
 
 /**
  * Locks an AFC client, done for thread safety stuff
- * 
+ *
  * @param client The AFC client connection to lock
  */
 static void afc_lock(afc_client_t client)
@@ -46,8 +47,8 @@ static void afc_lock(afc_client_t client)
 
 /**
  * Unlocks an AFC client, done for thread safety stuff.
- * 
- * @param client The AFC 
+ *
+ * @param client The AFC
  */
 static void afc_unlock(afc_client_t client)
 {
@@ -62,12 +63,12 @@ static void afc_unlock(afc_client_t client)
  * @param service_client A connected service client
  * @param client Pointer that will be set to a newly allocated afc_client_t
  *     upon successful return.
- * 
+ *
  * @return AFC_E_SUCCESS on success, AFC_E_INVALID_ARG if connection is
  *  invalid, or AFC_E_NO_MEM if there is a memory allocation problem.
  */
 
-afc_error_t afc_client_new_with_service_client(service_client_t service_client, afc_client_t *client)
+LIBIMOBILEDEVICE_API afc_error_t afc_client_new_with_service_client(service_client_t service_client, afc_client_t *client)
 {
 	if (!service_client)
 		return AFC_E_INVALID_ARG;
@@ -77,37 +78,23 @@ afc_error_t afc_client_new_with_service_client(service_client_t service_client,
 	client_loc->free_parent = 0;
 
 	/* allocate a packet */
-	client_loc->afc_packet = (AFCPacket *) malloc(sizeof(AFCPacket));
+	client_loc->packet_extra = 1024;
+	client_loc->afc_packet = (AFCPacket *) malloc(sizeof(AFCPacket) + client_loc->packet_extra);
 	if (!client_loc->afc_packet) {
 		free(client_loc);
 		return AFC_E_NO_MEM;
 	}
-
 	client_loc->afc_packet->packet_num = 0;
 	client_loc->afc_packet->entire_length = 0;
 	client_loc->afc_packet->this_length = 0;
 	memcpy(client_loc->afc_packet->magic, AFC_MAGIC, AFC_MAGIC_LEN);
-	client_loc->file_handle = 0;
-	client_loc->lock = 0;
 	mutex_init(&client_loc->mutex);
 
 	*client = client_loc;
 	return AFC_E_SUCCESS;
 }
 
-/**
- * Makes a connection to the AFC service on the device.
- * 
- * @param device The device to connect to.
- * @param service The service descriptor returned by lockdownd_start_service.
- * @param client Pointer that will be set to a newly allocated afc_client_t
- *     upon successful return.
- * 
- * @return AFC_E_SUCCESS on success, AFC_E_INVALID_ARG if device or service is
- *  invalid, AFC_E_MUX_ERROR if the connection cannot be established,
- *  or AFC_E_NO_MEM if there is a memory allocation problem.
- */
-afc_error_t afc_client_new(idevice_t device, lockdownd_service_descriptor_t service, afc_client_t * client)
+LIBIMOBILEDEVICE_API afc_error_t afc_client_new(idevice_t device, lockdownd_service_descriptor_t service, afc_client_t * client)
 {
 	if (!device || !service || service->port == 0)
 		return AFC_E_INVALID_ARG;
@@ -126,33 +113,14 @@ afc_error_t afc_client_new(idevice_t device, lockdownd_service_descriptor_t serv
 	return err;
 }
 
-/**
- * Starts a new AFC service on the specified device and connects to it.
- *
- * @param device The device to connect to.
- * @param client Pointer that will point to a newly allocated
- *     afc_client_t upon successful return. Must be freed using
- *     afc_client_free() after use.
- * @param label The label to use for communication. Usually the program name.
- *  Pass NULL to disable sending the label in requests to lockdownd.
- *
- * @return AFC_E_SUCCESS on success, or an AFC_E_* error
- *     code otherwise.
- */
-afc_error_t afc_client_start_service(idevice_t device, afc_client_t * client, const char* label)
+LIBIMOBILEDEVICE_API afc_error_t afc_client_start_service(idevice_t device, afc_client_t * client, const char* label)
 {
 	afc_error_t err = AFC_E_UNKNOWN_ERROR;
 	service_client_factory_start_service(device, AFC_SERVICE_NAME, (void**)client, label, SERVICE_CONSTRUCTOR(afc_client_new), &err);
 	return err;
 }
 
-/**
- * Frees up an AFC client. If the connection was created by the
- * client itself, the connection will be closed.
- * 
- * @param client The client to free.
- */
-afc_error_t afc_client_free(afc_client_t client)
+LIBIMOBILEDEVICE_API afc_error_t afc_client_free(afc_client_t client)
 {
 	if (!client || !client->afc_packet)
 		return AFC_E_INVALID_ARG;
@@ -169,7 +137,7 @@ afc_error_t afc_client_free(afc_client_t client)
 
 /**
  * Dispatches an AFC packet over a client.
- * 
+ *
  * @param client The client to send data through.
  * @param operation The operation to perform.
  * @param data The data to send together with the header.
@@ -180,7 +148,7 @@ afc_error_t afc_client_free(afc_client_t client)
  *
  * @return AFC_E_SUCCESS on success or an AFC_E_* error value.
  */
-static afc_error_t afc_dispatch_packet(afc_client_t client, uint64_t operation, const char *data, uint32_t data_length, const char* payload, uint32_t payload_length, uint32_t *bytes_sent)
+static afc_error_t afc_dispatch_packet(afc_client_t client, uint64_t operation, uint32_t data_length, const char* payload, uint32_t payload_length, uint32_t *bytes_sent)
 {
 	uint32_t sent = 0;
 
@@ -189,8 +157,6 @@ static afc_error_t afc_dispatch_packet(afc_client_t client, uint64_t operation,
 
 	*bytes_sent = 0;
 
-	if (!data || !data_length)
-		data_length = 0;
 	if (!payload || !payload_length)
 		payload_length = 0;
 
@@ -201,34 +167,26 @@ static afc_error_t afc_dispatch_packet(afc_client_t client, uint64_t operation,
 
 	debug_info("packet length = %i", client->afc_packet->this_length);
 
-	debug_buffer((char*)client->afc_packet, sizeof(AFCPacket));
-
-	/* send AFC packet header */
+	/* send AFC packet header and data */
 	AFCPacket_to_LE(client->afc_packet);
+	debug_buffer((char*)client->afc_packet, sizeof(AFCPacket) + data_length);
 	sent = 0;
-	service_send(client->parent, (void*)client->afc_packet, sizeof(AFCPacket), &sent);
+	service_send(client->parent, (void*)client->afc_packet, sizeof(AFCPacket) + data_length, &sent);
 	AFCPacket_from_LE(client->afc_packet);
 	*bytes_sent += sent;
-	if (sent < sizeof(AFCPacket)) {
-		return AFC_E_SUCCESS;
-	}
-
-	/* send AFC packet data (if there's data to send) */
-	sent = 0;
-	if (data_length > 0) {
-		debug_info("packet data follows");
-		debug_buffer(data, data_length);
-		service_send(client->parent, data, data_length, &sent);
-	}
-	*bytes_sent += sent;
-	if (sent < data_length) {
+	if (sent < sizeof(AFCPacket) + data_length) {
 		return AFC_E_SUCCESS;
 	}
 
 	sent = 0;
 	if (payload_length > 0) {
-		debug_info("packet payload follows");
-		debug_buffer(payload, payload_length);
+		if (payload_length > 256) {
+			debug_info("packet payload follows (256/%u)", payload_length);
+			debug_buffer(payload, 256);
+		} else {
+			debug_info("packet payload follows");
+			debug_buffer(payload, payload_length);
+		}
 		service_send(client->parent, payload, payload_length, &sent);
 	}
 	*bytes_sent += sent;
@@ -241,11 +199,11 @@ static afc_error_t afc_dispatch_packet(afc_client_t client, uint64_t operation,
 
 /**
  * Receives data through an AFC client and sets a variable to the received data.
- * 
+ *
  * @param client The client to receive data on.
  * @param bytes The char* to point to the newly-received data.
  * @param bytes_recv How much data was received.
- * 
+ *
  * @return AFC_E_SUCCESS on success or an AFC_E_* error value.
  */
 static afc_error_t afc_receive_data(afc_client_t client, char **bytes, uint32_t *bytes_recv)
@@ -342,8 +300,13 @@ static afc_error_t afc_receive_data(afc_client_t client, char **bytes, uint32_t
 	}
 
 	debug_info("packet data size = %i", current_count);
-	debug_info("packet data follows");
-	debug_buffer(dump_here, current_count);
+	if (current_count > 256) {
+		debug_info("packet data follows (256/%u)", current_count);
+		debug_buffer(dump_here, 256);
+	} else {
+		debug_info("packet data follows");
+		debug_buffer(dump_here, current_count);
+	}
 
 	/* check operation types */
 	if (header.operation == AFC_OP_STATUS) {
@@ -432,29 +395,42 @@ static char **make_strings_list(char *tokens, uint32_t length)
 	return list;
 }
 
-/**
- * Gets a directory listing of the directory requested.
- * 
- * @param client The client to get a directory listing from.
- * @param dir The directory to list. (must be a fully-qualified path)
- * @param list A char list of files in that directory, terminated by an empty
- *         string or NULL if there was an error.
- * 
- * @return AFC_E_SUCCESS on success or an AFC_E_* error value.
- */
-afc_error_t afc_read_directory(afc_client_t client, const char *dir, char ***list)
+static int _afc_check_packet_buffer(afc_client_t client, uint32_t data_len)
+{
+	if (data_len > client->packet_extra) {
+		client->packet_extra = (data_len & ~8) + 8;
+		AFCPacket* newpkt = (AFCPacket*)realloc(client->afc_packet, sizeof(AFCPacket) + client->packet_extra);
+		if (!newpkt) {
+			return -1;
+		}
+		client->afc_packet = newpkt;
+	}
+	return 0;
+}
+
+#define AFC_PACKET_DATA_PTR ((char*)client->afc_packet + sizeof(AFCPacket))
+
+LIBIMOBILEDEVICE_API afc_error_t afc_read_directory(afc_client_t client, const char *path, char ***directory_information)
 {
 	uint32_t bytes = 0;
 	char *data = NULL, **list_loc = NULL;
 	afc_error_t ret = AFC_E_UNKNOWN_ERROR;
 
-	if (!client || !dir || !list || (list && *list))
+	if (!client || !path || !directory_information || (directory_information && *directory_information))
 		return AFC_E_INVALID_ARG;
 
 	afc_lock(client);
 
+	uint32_t data_len = (uint32_t)strlen(path)+1;
+	if (_afc_check_packet_buffer(client, data_len) < 0) {
+		afc_unlock(client);
+		debug_info("Failed to realloc packet buffer");
+		return AFC_E_NO_MEM;
+	}
+
 	/* Send the command */
-	ret = afc_dispatch_packet(client, AFC_OP_READ_DIR, dir, strlen(dir)+1, NULL, 0, &bytes);
+	memcpy(AFC_PACKET_DATA_PTR, path, data_len);
+	ret = afc_dispatch_packet(client, AFC_OP_READ_DIR, data_len, NULL, 0, &bytes);
 	if (ret != AFC_E_SUCCESS) {
 		afc_unlock(client);
 		return AFC_E_NOT_ENOUGH_DATA;
@@ -473,35 +449,24 @@ afc_error_t afc_read_directory(afc_client_t client, const char *dir, char ***lis
 		free(data);
 
 	afc_unlock(client);
-	*list = list_loc;
+	*directory_information = list_loc;
 
 	return ret;
 }
 
-/**
- * Get device information for a connected client. The device information
- * returned is the device model as well as the free space, the total capacity
- * and blocksize on the accessed disk partition.
- * 
- * @param client The client to get device info for.
- * @param infos A char ** list of parameters as given by AFC or NULL if there
- *  was an error.
- * 
- * @return AFC_E_SUCCESS on success or an AFC_E_* error value.
- */
-afc_error_t afc_get_device_info(afc_client_t client, char ***infos)
+LIBIMOBILEDEVICE_API afc_error_t afc_get_device_info(afc_client_t client, char ***device_information)
 {
 	uint32_t bytes = 0;
 	char *data = NULL, **list = NULL;
 	afc_error_t ret = AFC_E_UNKNOWN_ERROR;
 
-	if (!client || !infos)
+	if (!client || !device_information)
 		return AFC_E_INVALID_ARG;
 
 	afc_lock(client);
 
 	/* Send the command */
-	ret = afc_dispatch_packet(client, AFC_OP_GET_DEVINFO, NULL, 0, NULL, 0, &bytes);
+	ret = afc_dispatch_packet(client, AFC_OP_GET_DEVINFO, 0, NULL, 0, &bytes);
 	if (ret != AFC_E_SUCCESS) {
 		afc_unlock(client);
 		return AFC_E_NOT_ENOUGH_DATA;
@@ -521,23 +486,12 @@ afc_error_t afc_get_device_info(afc_client_t client, char ***infos)
 
 	afc_unlock(client);
 
-	*infos = list;
+	*device_information = list;
 
 	return ret;
 }
 
-/**
- * Get a specific key of the device info list for a client connection.
- * Known key values are: Model, FSTotalBytes, FSFreeBytes and FSBlockSize.
- * This is a helper function for afc_get_device_info().
- *
- * @param client The client to get device info for.
- * @param key The key to get the value of.
- * @param value The value for the key if successful or NULL otherwise.
- *
- * @return AFC_E_SUCCESS on success or an AFC_E_* error value.
- */
-afc_error_t afc_get_device_info_key(afc_client_t client, const char *key, char **value)
+LIBIMOBILEDEVICE_API afc_error_t afc_get_device_info_key(afc_client_t client, const char *key, char **value)
 {
 	afc_error_t ret = AFC_E_INTERNAL_ERROR;
 	char **kvps, **ptr;
@@ -564,15 +518,7 @@ afc_error_t afc_get_device_info_key(afc_client_t client, const char *key, char *
 	return ret;
 }
 
-/**
- * Deletes a file or directory.
- * 
- * @param client The client to use.
- * @param path The path to delete. (must be a fully-qualified path)
- * 
- * @return AFC_E_SUCCESS on success or an AFC_E_* error value.
- */
-afc_error_t afc_remove_path(afc_client_t client, const char *path)
+LIBIMOBILEDEVICE_API afc_error_t afc_remove_path(afc_client_t client, const char *path)
 {
 	uint32_t bytes = 0;
 	afc_error_t ret = AFC_E_UNKNOWN_ERROR;
@@ -582,8 +528,16 @@ afc_error_t afc_remove_path(afc_client_t client, const char *path)
 
 	afc_lock(client);
 
+	uint32_t data_len = (uint32_t)strlen(path)+1;
+	if (_afc_check_packet_buffer(client, data_len) < 0) {
+		afc_unlock(client);
+		debug_info("Failed to realloc packet buffer");
+		return AFC_E_NO_MEM;
+	}
+
 	/* Send command */
-	ret = afc_dispatch_packet(client, AFC_OP_REMOVE_PATH, path, strlen(path)+1, NULL, 0, &bytes);
+	memcpy(AFC_PACKET_DATA_PTR, path, data_len);
+	ret = afc_dispatch_packet(client, AFC_OP_REMOVE_PATH, data_len, NULL, 0, &bytes);
 	if (ret != AFC_E_SUCCESS) {
 		afc_unlock(client);
 		return AFC_E_NOT_ENOUGH_DATA;
@@ -600,32 +554,30 @@ afc_error_t afc_remove_path(afc_client_t client, const char *path)
 	return ret;
 }
 
-/**
- * Renames a file or directory on the device.
- * 
- * @param client The client to have rename.
- * @param from The name to rename from. (must be a fully-qualified path)
- * @param to The new name. (must also be a fully-qualified path)
- * 
- * @return AFC_E_SUCCESS on success or an AFC_E_* error value.
- */
-afc_error_t afc_rename_path(afc_client_t client, const char *from, const char *to)
+LIBIMOBILEDEVICE_API afc_error_t afc_rename_path(afc_client_t client, const char *from, const char *to)
 {
-	char *send = (char *) malloc(sizeof(char) * (strlen(from) + strlen(to) + 1 + sizeof(uint32_t)));
-	uint32_t bytes = 0;
-	afc_error_t ret = AFC_E_UNKNOWN_ERROR;
-
 	if (!client || !from || !to || !client->afc_packet || !client->parent)
 		return AFC_E_INVALID_ARG;
 
+	uint32_t bytes = 0;
+	afc_error_t ret = AFC_E_UNKNOWN_ERROR;
+
+	size_t from_len = strlen(from);
+	size_t to_len = strlen(to);
+
 	afc_lock(client);
 
-	/* Send command */
-	memcpy(send, from, strlen(from) + 1);
-	memcpy(send + strlen(from) + 1, to, strlen(to) + 1);
-	ret = afc_dispatch_packet(client, AFC_OP_RENAME_PATH, send, strlen(to)+1 + strlen(from)+1, NULL, 0, &bytes);
-	free(send);
+	uint32_t data_len = (uint32_t)(from_len+1 + to_len+1);
+	if (_afc_check_packet_buffer(client, data_len) < 0) {
+		afc_unlock(client);
+		debug_info("Failed to realloc packet buffer");
+		return AFC_E_NO_MEM;
+	}
 
+	/* Send command */
+	memcpy(AFC_PACKET_DATA_PTR, from, from_len+1);
+	memcpy(AFC_PACKET_DATA_PTR + from_len+1, to, to_len+1);
+	ret = afc_dispatch_packet(client, AFC_OP_RENAME_PATH, data_len, NULL, 0, &bytes);
 	if (ret != AFC_E_SUCCESS) {
 		afc_unlock(client);
 		return AFC_E_NOT_ENOUGH_DATA;
@@ -638,16 +590,7 @@ afc_error_t afc_rename_path(afc_client_t client, const char *from, const char *t
 	return ret;
 }
 
-/**
- * Creates a directory on the device.
- * 
- * @param client The client to use to make a directory.
- * @param dir The directory's path. (must be a fully-qualified path, I assume
- *        all other mkdir restrictions apply as well)
- *
- * @return AFC_E_SUCCESS on success or an AFC_E_* error value.
- */
-afc_error_t afc_make_directory(afc_client_t client, const char *dir)
+LIBIMOBILEDEVICE_API afc_error_t afc_make_directory(afc_client_t client, const char *path)
 {
 	uint32_t bytes = 0;
 	afc_error_t ret = AFC_E_UNKNOWN_ERROR;
@@ -657,8 +600,16 @@ afc_error_t afc_make_directory(afc_client_t client, const char *dir)
 
 	afc_lock(client);
 
+	uint32_t data_len = (uint32_t)strlen(path)+1;
+	if (_afc_check_packet_buffer(client, data_len) < 0) {
+		afc_unlock(client);
+		debug_info("Failed to realloc packet buffer");
+		return AFC_E_NO_MEM;
+	}
+
 	/* Send command */
-	ret = afc_dispatch_packet(client, AFC_OP_MAKE_DIR, dir, strlen(dir)+1, NULL, 0, &bytes);
+	memcpy(AFC_PACKET_DATA_PTR, path, data_len);
+	ret = afc_dispatch_packet(client, AFC_OP_MAKE_DIR, data_len, NULL, 0, &bytes);
 	if (ret != AFC_E_SUCCESS) {
 		afc_unlock(client);
 		return AFC_E_NOT_ENOUGH_DATA;
@@ -671,30 +622,29 @@ afc_error_t afc_make_directory(afc_client_t client, const char *dir)
 	return ret;
 }
 
-/**
- * Gets information about a specific file.
- * 
- * @param client The client to use to get the information of the file.
- * @param path The fully-qualified path to the file. 
- * @param infolist Pointer to a buffer that will be filled with a NULL-terminated
- *                 list of strings with the file information.
- *                 Set to NULL before calling this function.
- * 
- * @return AFC_E_SUCCESS on success or an AFC_E_* error value.
- */
-afc_error_t afc_get_file_info(afc_client_t client, const char *path, char ***infolist)
+LIBIMOBILEDEVICE_API afc_error_t afc_get_file_info(afc_client_t client, const char *path, char ***file_information)
 {
 	char *received = NULL;
 	uint32_t bytes = 0;
 	afc_error_t ret = AFC_E_UNKNOWN_ERROR;
 
-	if (!client || !path || !infolist)
+	if (!client || !path || !file_information)
 		return AFC_E_INVALID_ARG;
 
 	afc_lock(client);
 
+	uint32_t data_len = (uint32_t)strlen(path)+1;
+	if (_afc_check_packet_buffer(client, data_len) < 0) {
+		afc_unlock(client);
+		debug_info("Failed to realloc packet buffer");
+		return AFC_E_NO_MEM;
+	}
+
+	debug_info("We got %p and %p", client->afc_packet, AFC_PACKET_DATA_PTR);
+
 	/* Send command */
-	ret = afc_dispatch_packet(client, AFC_OP_GET_FILE_INFO, path, strlen(path)+1, NULL, 0, &bytes);
+	memcpy(AFC_PACKET_DATA_PTR, path, data_len);
+	ret = afc_dispatch_packet(client, AFC_OP_GET_FILE_INFO, data_len, NULL, 0, &bytes);
 	if (ret != AFC_E_SUCCESS) {
 		afc_unlock(client);
 		return AFC_E_NOT_ENOUGH_DATA;
@@ -703,7 +653,7 @@ afc_error_t afc_get_file_info(afc_client_t client, const char *path, char ***inf
 	/* Receive data */
 	ret = afc_receive_data(client, &received, &bytes);
 	if (received) {
-		*infolist = make_strings_list(received, bytes);
+		*file_information = make_strings_list(received, bytes);
 		free(received);
 	}
 
@@ -712,49 +662,39 @@ afc_error_t afc_get_file_info(afc_client_t client, const char *path, char ***inf
 	return ret;
 }
 
-/**
- * Opens a file on the device.
- * 
- * @param client The client to use to open the file. 
- * @param filename The file to open. (must be a fully-qualified path)
- * @param file_mode The mode to use to open the file. Can be AFC_FILE_READ or
- * 		    AFC_FILE_WRITE; the former lets you read and write,
- * 		    however, and the second one will *create* the file,
- * 		    destroying anything previously there.
- * @param handle Pointer to a uint64_t that will hold the handle of the file
- * 
- * @return AFC_E_SUCCESS on success or an AFC_E_* error value.
- */
-idevice_error_t
-afc_file_open(afc_client_t client, const char *filename,
-					 afc_file_mode_t file_mode, uint64_t *handle)
+LIBIMOBILEDEVICE_API afc_error_t afc_file_open(afc_client_t client, const char *filename, afc_file_mode_t file_mode, uint64_t *handle)
 {
-	uint64_t file_mode_loc = htole64(file_mode);
+	if (!client || !client->parent || !client->afc_packet)
+		return AFC_E_INVALID_ARG;
+
+	//uint64_t file_mode_loc = htole64(file_mode);
 	uint32_t bytes = 0;
-	char *data = (char *) malloc(sizeof(char) * (8 + strlen(filename) + 1));
 	afc_error_t ret = AFC_E_UNKNOWN_ERROR;
 
 	/* set handle to 0 so in case an error occurs, the handle is invalid */
 	*handle = 0;
 
-	if (!client || !client->parent || !client->afc_packet)
-		return AFC_E_INVALID_ARG;
-
 	afc_lock(client);
 
-	/* Send command */
-	memcpy(data, &file_mode_loc, 8);
-	memcpy(data + 8, filename, strlen(filename));
-	data[8 + strlen(filename)] = '\0';
-	ret = afc_dispatch_packet(client, AFC_OP_FILE_OPEN, data, 8 + strlen(filename) + 1, NULL, 0, &bytes);
-	free(data);
+	uint32_t data_len = (uint32_t)(strlen(filename)+1 + 8);
+	if (_afc_check_packet_buffer(client, data_len) < 0) {
+		afc_unlock(client);
+		debug_info("Failed to realloc packet buffer");
+		return AFC_E_NO_MEM;
+	}
 
+	/* Send command */
+	//memcpy(AFC_PACKET_DATA_PTR, &file_mode_loc, 8);
+	*(uint64_t*)(AFC_PACKET_DATA_PTR) = htole64(file_mode);
+	memcpy(AFC_PACKET_DATA_PTR + 8, filename, data_len-8);
+	ret = afc_dispatch_packet(client, AFC_OP_FILE_OPEN, data_len, NULL, 0, &bytes);
 	if (ret != AFC_E_SUCCESS) {
 		debug_info("Didn't receive a response to the command");
 		afc_unlock(client);
 		return AFC_E_NOT_ENOUGH_DATA;
 	}
 	/* Receive the data */
+	char* data = NULL;
 	ret = afc_receive_data(client, &data, &bytes);
 	if ((ret == AFC_E_SUCCESS) && (bytes > 0) && data) {
 		afc_unlock(client);
@@ -764,6 +704,8 @@ afc_file_open(afc_client_t client, const char *filename,
 		free(data);
 		return ret;
 	}
+	/* in case memory was allocated but no data received or an error occurred */
+	free(data);
 
 	debug_info("Didn't get any further data");
 
@@ -772,39 +714,29 @@ afc_file_open(afc_client_t client, const char *filename,
 	return ret;
 }
 
-/**
- * Attempts to the read the given number of bytes from the given file.
- * 
- * @param client The relevant AFC client
- * @param handle File handle of a previously opened file
- * @param data The pointer to the memory region to store the read data
- * @param length The number of bytes to read
- * @param bytes_read The number of bytes actually read.
- *
- * @return AFC_E_SUCCESS on success or an AFC_E_* error value.
- */
-idevice_error_t
-afc_file_read(afc_client_t client, uint64_t handle, char *data, uint32_t length, uint32_t *bytes_read)
+LIBIMOBILEDEVICE_API afc_error_t afc_file_read(afc_client_t client, uint64_t handle, char *data, uint32_t length, uint32_t *bytes_read)
 {
 	char *input = NULL;
 	uint32_t current_count = 0, bytes_loc = 0;
+	struct readinfo {
+		uint64_t handle;
+		uint64_t size;
+	};
 	afc_error_t ret = AFC_E_SUCCESS;
 
 	if (!client || !client->afc_packet || !client->parent || handle == 0)
 		return AFC_E_INVALID_ARG;
 	debug_info("called for length %i", length);
 
+	//uint32_t data_len = 8 + 8;
+
 	afc_lock(client);
 
 	/* Send the read command */
-	struct {
-		uint64_t handle;
-		uint64_t size;
-	} readinfo;
-	readinfo.handle = handle;
-	readinfo.size = htole64(length);
-	ret = afc_dispatch_packet(client, AFC_OP_READ, (const char*)&readinfo, sizeof(readinfo), NULL, 0, &bytes_loc);
-
+	struct readinfo* readinfo = (struct readinfo*)(AFC_PACKET_DATA_PTR);
+	readinfo->handle = handle;
+	readinfo->size = htole64(length);
+	ret = afc_dispatch_packet(client, AFC_OP_FILE_READ, sizeof(struct readinfo), NULL, 0, &bytes_loc);
 	if (ret != AFC_E_SUCCESS) {
 		afc_unlock(client);
 		return AFC_E_NOT_ENOUGH_DATA;
@@ -837,19 +769,7 @@ afc_file_read(afc_client_t client, uint64_t handle, char *data, uint32_t length,
 	return ret;
 }
 
-/**
- * Writes a given number of bytes to a file.
- * 
- * @param client The client to use to write to the file.
- * @param handle File handle of previously opened file. 
- * @param data The data to write to the file.
- * @param length How much data to write.
- * @param bytes_written The number of bytes actually written to the file.
- * 
- * @return AFC_E_SUCCESS on success or an AFC_E_* error value.
- */
-idevice_error_t
-afc_file_write(afc_client_t client, uint64_t handle, const char *data, uint32_t length, uint32_t *bytes_written)
+LIBIMOBILEDEVICE_API afc_error_t afc_file_write(afc_client_t client, uint64_t handle, const char *data, uint32_t length, uint32_t *bytes_written)
 {
 	uint32_t current_count = 0;
 	uint32_t bytes_loc = 0;
@@ -858,11 +778,14 @@ afc_file_write(afc_client_t client, uint64_t handle, const char *data, uint32_t
 	if (!client || !client->afc_packet || !client->parent || !bytes_written || (handle == 0))
 		return AFC_E_INVALID_ARG;
 
+	uint32_t data_len = 8;
+
 	afc_lock(client);
 
 	debug_info("Write length: %i", length);
 
-	ret = afc_dispatch_packet(client, AFC_OP_WRITE, (const char*)&handle, 8, data, length, &bytes_loc);
+	*(uint64_t*)(AFC_PACKET_DATA_PTR) = handle;
+	ret = afc_dispatch_packet(client, AFC_OP_FILE_WRITE, data_len, data, length, &bytes_loc);
 
 	current_count += bytes_loc - (sizeof(AFCPacket) + 8);
 
@@ -881,13 +804,7 @@ afc_file_write(afc_client_t client, uint64_t handle, const char *data, uint32_t
 	return ret;
 }
 
-/**
- * Closes a file on the device.
- * 
- * @param client The client to close the file with.
- * @param handle File handle of a previously opened file.
- */
-afc_error_t afc_file_close(afc_client_t client, uint64_t handle)
+LIBIMOBILEDEVICE_API afc_error_t afc_file_close(afc_client_t client, uint64_t handle)
 {
 	uint32_t bytes = 0;
 	afc_error_t ret = AFC_E_UNKNOWN_ERROR;
@@ -895,12 +812,15 @@ afc_error_t afc_file_close(afc_client_t client, uint64_t handle)
 	if (!client || (handle == 0))
 		return AFC_E_INVALID_ARG;
 
+	uint32_t data_len = 8;
+
 	afc_lock(client);
 
 	debug_info("File handle %i", handle);
 
 	/* Send command */
-	ret = afc_dispatch_packet(client, AFC_OP_FILE_CLOSE, (const char*)&handle, 8, NULL, 0, &bytes);
+	*(uint64_t*)(AFC_PACKET_DATA_PTR) = handle;
+	ret = afc_dispatch_packet(client, AFC_OP_FILE_CLOSE, data_len, NULL, 0, &bytes);
 
 	if (ret != AFC_E_SUCCESS) {
 		afc_unlock(client);
@@ -915,25 +835,13 @@ afc_error_t afc_file_close(afc_client_t client, uint64_t handle)
 	return ret;
 }
 
-/**
- * Locks or unlocks a file on the device. 
- *
- * makes use of flock on the device, see
- * http://developer.apple.com/documentation/Darwin/Reference/ManPages/man2/flock.2.html
- *
- * @param client The client to lock the file with.
- * @param handle File handle of a previously opened file.
- * @param operation the lock or unlock operation to perform, this is one of
- *        AFC_LOCK_SH (shared lock), AFC_LOCK_EX (exclusive lock),
- *        or AFC_LOCK_UN (unlock).
- */
-afc_error_t afc_file_lock(afc_client_t client, uint64_t handle, afc_lock_op_t operation)
+LIBIMOBILEDEVICE_API afc_error_t afc_file_lock(afc_client_t client, uint64_t handle, afc_lock_op_t operation)
 {
 	uint32_t bytes = 0;
-	struct {
+	struct lockinfo {
 		uint64_t handle;
 		uint64_t op;
-	} lockinfo;
+	};
 	afc_error_t ret = AFC_E_UNKNOWN_ERROR;
 
 	if (!client || (handle == 0))
@@ -944,10 +852,10 @@ afc_error_t afc_file_lock(afc_client_t client, uint64_t handle, afc_lock_op_t op
 	debug_info("file handle %i", handle);
 
 	/* Send command */
-	lockinfo.handle = handle;
-	lockinfo.op = htole64(operation);
-	ret = afc_dispatch_packet(client, AFC_OP_FILE_LOCK, (const char*)&lockinfo, sizeof(lockinfo), NULL, 0, &bytes);
-
+	struct lockinfo* lockinfo = (struct lockinfo*)(AFC_PACKET_DATA_PTR);
+	lockinfo->handle = handle;
+	lockinfo->op = htole64(operation);
+	ret = afc_dispatch_packet(client, AFC_OP_FILE_LOCK, sizeof(struct lockinfo), NULL, 0, &bytes);
 	if (ret != AFC_E_SUCCESS) {
 		afc_unlock(client);
 		debug_info("could not send lock command");
@@ -961,24 +869,14 @@ afc_error_t afc_file_lock(afc_client_t client, uint64_t handle, afc_lock_op_t op
 	return ret;
 }
 
-/**
- * Seeks to a given position of a pre-opened file on the device. 
- * 
- * @param client The client to use to seek to the position.
- * @param handle File handle of a previously opened.
- * @param offset Seek offset.
- * @param whence Seeking direction, one of SEEK_SET, SEEK_CUR, or SEEK_END.
- * 
- * @return AFC_E_SUCCESS on success or an AFC_E_* error value.
- */
-afc_error_t afc_file_seek(afc_client_t client, uint64_t handle, int64_t offset, int whence)
+LIBIMOBILEDEVICE_API afc_error_t afc_file_seek(afc_client_t client, uint64_t handle, int64_t offset, int whence)
 {
 	uint32_t bytes = 0;
-	struct {
+	struct seekinfo {
 		uint64_t handle;
 		uint64_t whence;
 		int64_t offset;
-	} seekinfo;
+	};
 	afc_error_t ret = AFC_E_UNKNOWN_ERROR;
 
 	if (!client || (handle == 0))
@@ -987,10 +885,11 @@ afc_error_t afc_file_seek(afc_client_t client, uint64_t handle, int64_t offset,
 	afc_lock(client);
 
 	/* Send the command */
-	seekinfo.handle = handle;
-	seekinfo.whence = htole64(whence);
-	seekinfo.offset = (int64_t)htole64(offset);
-	ret = afc_dispatch_packet(client, AFC_OP_FILE_SEEK, (const char*)&seekinfo, sizeof(seekinfo), NULL, 0, &bytes);
+	struct seekinfo* seekinfo = (struct seekinfo*)(AFC_PACKET_DATA_PTR);
+	seekinfo->handle = handle;
+	seekinfo->whence = htole64(whence);
+	seekinfo->offset = (int64_t)htole64(offset);
+	ret = afc_dispatch_packet(client, AFC_OP_FILE_SEEK, sizeof(struct seekinfo), NULL, 0, &bytes);
 
 	if (ret != AFC_E_SUCCESS) {
 		afc_unlock(client);
@@ -1004,29 +903,22 @@ afc_error_t afc_file_seek(afc_client_t client, uint64_t handle, int64_t offset,
 	return ret;
 }
 
-/**
- * Returns current position in a pre-opened file on the device.
- * 
- * @param client The client to use.
- * @param handle File handle of a previously opened file.
- * @param position Position in bytes of indicator
- * 
- * @return AFC_E_SUCCESS on success or an AFC_E_* error value.
- */
-afc_error_t afc_file_tell(afc_client_t client, uint64_t handle, uint64_t *position)
+LIBIMOBILEDEVICE_API afc_error_t afc_file_tell(afc_client_t client, uint64_t handle, uint64_t *position)
 {
-	char *buffer = (char *) malloc(sizeof(char) * 8);
+	char *buffer = NULL;
 	uint32_t bytes = 0;
 	afc_error_t ret = AFC_E_UNKNOWN_ERROR;
 
 	if (!client || (handle == 0))
 		return AFC_E_INVALID_ARG;
 
+	uint32_t data_len = 8;
+
 	afc_lock(client);
 
 	/* Send the command */
-	ret = afc_dispatch_packet(client, AFC_OP_FILE_TELL, (const char*)&handle, 8, NULL, 0, &bytes);
-
+	*(uint64_t*)(AFC_PACKET_DATA_PTR) = handle;
+	ret = afc_dispatch_packet(client, AFC_OP_FILE_TELL, data_len, NULL, 0, &bytes);
 	if (ret != AFC_E_SUCCESS) {
 		afc_unlock(client);
 		return AFC_E_NOT_ENOUGH_DATA;
@@ -1039,33 +931,20 @@ afc_error_t afc_file_tell(afc_client_t client, uint64_t handle, uint64_t *positi
 		memcpy(position, buffer, sizeof(uint64_t));
 		*position = le64toh(*position);
 	}
-	if (buffer)
-		free(buffer);
+	free(buffer);
 
 	afc_unlock(client);
 
 	return ret;
 }
 
-/**
- * Sets the size of a file on the device.
- * 
- * @param client The client to use to set the file size.
- * @param handle File handle of a previously opened file.
- * @param newsize The size to set the file to. 
- * 
- * @return AFC_E_SUCCESS on success or an AFC_E_* error value.
- * 
- * @note This function is more akin to ftruncate than truncate, and truncate
- *       calls would have to open the file before calling this, sadly.
- */
-afc_error_t afc_file_truncate(afc_client_t client, uint64_t handle, uint64_t newsize)
+LIBIMOBILEDEVICE_API afc_error_t afc_file_truncate(afc_client_t client, uint64_t handle, uint64_t newsize)
 {
 	uint32_t bytes = 0;
-	struct {
+	struct truncinfo {
 		uint64_t handle;
 		uint64_t newsize;
-	} truncinfo;
+	};
 	afc_error_t ret = AFC_E_UNKNOWN_ERROR;
 
 	if (!client || (handle == 0))
@@ -1074,9 +953,10 @@ afc_error_t afc_file_truncate(afc_client_t client, uint64_t handle, uint64_t new
 	afc_lock(client);
 
 	/* Send command */
-	truncinfo.handle = handle;
-	truncinfo.newsize = htole64(newsize);
-	ret = afc_dispatch_packet(client, AFC_OP_FILE_SET_SIZE, (const char*)&truncinfo, sizeof(truncinfo), NULL, 0, &bytes);
+	struct truncinfo* truncinfo = (struct truncinfo*)(AFC_PACKET_DATA_PTR);
+	truncinfo->handle = handle;
+	truncinfo->newsize = htole64(newsize);
+	ret = afc_dispatch_packet(client, AFC_OP_FILE_SET_SIZE, sizeof(struct truncinfo), NULL, 0, &bytes);
 
 	if (ret != AFC_E_SUCCESS) {
 		afc_unlock(client);
@@ -1090,33 +970,27 @@ afc_error_t afc_file_truncate(afc_client_t client, uint64_t handle, uint64_t new
 	return ret;
 }
 
-/**
- * Sets the size of a file on the device without prior opening it.
- * 
- * @param client The client to use to set the file size.
- * @param path The path of the file to be truncated.
- * @param newsize The size to set the file to. 
- * 
- * @return AFC_E_SUCCESS on success or an AFC_E_* error value.
- */
-afc_error_t afc_truncate(afc_client_t client, const char *path, uint64_t newsize)
+LIBIMOBILEDEVICE_API afc_error_t afc_truncate(afc_client_t client, const char *path, uint64_t newsize)
 {
-	char *send = (char *) malloc(sizeof(char) * (strlen(path) + 1 + 8));
-	uint32_t bytes = 0;
-	uint64_t size_requested = htole64(newsize);
-	afc_error_t ret = AFC_E_UNKNOWN_ERROR;
-
 	if (!client || !path || !client->afc_packet || !client->parent)
 		return AFC_E_INVALID_ARG;
 
+	uint32_t bytes = 0;
+	afc_error_t ret = AFC_E_UNKNOWN_ERROR;
+
 	afc_lock(client);
 
-	/* Send command */
-	memcpy(send, &size_requested, 8);
-	memcpy(send + 8, path, strlen(path) + 1);
-	ret = afc_dispatch_packet(client, AFC_OP_TRUNCATE, send, 8 + strlen(path) + 1, NULL, 0, &bytes);
-	free(send);
+	uint32_t data_len = 8 + (uint32_t)(strlen(path)+1);
+	if (_afc_check_packet_buffer(client, data_len) < 0) {
+		afc_unlock(client);
+		debug_info("Failed to realloc packet buffer");
+		return AFC_E_NO_MEM;
+	}
 
+	/* Send command */
+	*(uint64_t*)(AFC_PACKET_DATA_PTR) = htole64(newsize);
+	memcpy(AFC_PACKET_DATA_PTR + 8, path, data_len-8);
+	ret = afc_dispatch_packet(client, AFC_OP_TRUNCATE, data_len, NULL, 0, &bytes);
 	if (ret != AFC_E_SUCCESS) {
 		afc_unlock(client);
 		return AFC_E_NOT_ENOUGH_DATA;
@@ -1129,38 +1003,36 @@ afc_error_t afc_truncate(afc_client_t client, const char *path, uint64_t newsize
 	return ret;
 }
 
-/**
- * Creates a hard link or symbolic link on the device. 
- * 
- * @param client The client to use for making a link
- * @param linktype 1 = hard link, 2 = symlink
- * @param target The file to be linked.
- * @param linkname The name of link.
- * 
- * @return AFC_E_SUCCESS on success or an AFC_E_* error value.
- */
-afc_error_t afc_make_link(afc_client_t client, afc_link_type_t linktype, const char *target, const char *linkname)
+LIBIMOBILEDEVICE_API afc_error_t afc_make_link(afc_client_t client, afc_link_type_t linktype, const char *target, const char *linkname)
 {
-	char *send = (char *) malloc(sizeof(char) * (strlen(target)+1 + strlen(linkname)+1 + 8));
+	if (!client || !target || !linkname || !client->afc_packet || !client->parent)
+		return AFC_E_INVALID_ARG;
+
 	uint32_t bytes = 0;
 	uint64_t type = htole64(linktype);
 	afc_error_t ret = AFC_E_UNKNOWN_ERROR;
 
-	if (!client || !target || !linkname || !client->afc_packet || !client->parent)
-		return AFC_E_INVALID_ARG;
+	size_t target_len = strlen(target);
+	size_t link_len = strlen(linkname);
 
 	afc_lock(client);
 
+	uint32_t data_len = 8 + target_len + 1 + link_len + 1;
+	if (_afc_check_packet_buffer(client, data_len) < 0) {
+		afc_unlock(client);
+		debug_info("Failed to realloc packet buffer");
+		return AFC_E_NO_MEM;
+	}
+
 	debug_info("link type: %lld", type);
-	debug_info("target: %s, length:%d", target, strlen(target));
-	debug_info("linkname: %s, length:%d", linkname, strlen(linkname));
+	debug_info("target: %s, length:%d", target, target_len);
+	debug_info("linkname: %s, length:%d", linkname, link_len);
 
 	/* Send command */
-	memcpy(send, &type, 8);
-	memcpy(send + 8, target, strlen(target) + 1);
-	memcpy(send + 8 + strlen(target) + 1, linkname, strlen(linkname) + 1);
-	ret = afc_dispatch_packet(client, AFC_OP_MAKE_LINK, send, 8 + strlen(linkname) + 1 + strlen(target) + 1, NULL, 0, &bytes);
-	free(send);
+	*(uint64_t*)(AFC_PACKET_DATA_PTR) = htole64(linktype);
+	memcpy(AFC_PACKET_DATA_PTR + 8, target, target_len + 1);
+	memcpy(AFC_PACKET_DATA_PTR + 8 + target_len + 1, linkname, link_len + 1);
+	ret = afc_dispatch_packet(client, AFC_OP_MAKE_LINK, data_len, NULL, 0, &bytes);
 	if (ret != AFC_E_SUCCESS) {
 		afc_unlock(client);
 		return AFC_E_NOT_ENOUGH_DATA;
@@ -1173,20 +1045,42 @@ afc_error_t afc_make_link(afc_client_t client, afc_link_type_t linktype, const c
 	return ret;
 }
 
-/**
- * Sets the modification time of a file on the device.
- * 
- * @param client The client to use to set the file size.
- * @param path Path of the file for which the modification time should be set.
- * @param mtime The modification time to set in nanoseconds since epoch.
- * 
- * @return AFC_E_SUCCESS on success or an AFC_E_* error value.
- */
-afc_error_t afc_set_file_time(afc_client_t client, const char *path, uint64_t mtime)
+LIBIMOBILEDEVICE_API afc_error_t afc_set_file_time(afc_client_t client, const char *path, uint64_t mtime)
+{
+	if (!client || !path || !client->afc_packet || !client->parent)
+		return AFC_E_INVALID_ARG;
+
+	uint32_t bytes = 0;
+	afc_error_t ret = AFC_E_UNKNOWN_ERROR;
+
+	afc_lock(client);
+
+	uint32_t data_len = 8 + strlen(path) + 1;
+	if (_afc_check_packet_buffer(client, data_len) < 0) {
+		afc_unlock(client);
+		debug_info("Failed to realloc packet buffer");
+		return AFC_E_NO_MEM;
+	}
+
+	/* Send command */
+	*(uint64_t*)(AFC_PACKET_DATA_PTR) = htole64(mtime);
+	memcpy(AFC_PACKET_DATA_PTR + 8, path, data_len-8);
+	ret = afc_dispatch_packet(client, AFC_OP_SET_FILE_MOD_TIME, data_len, NULL, 0, &bytes);
+	if (ret != AFC_E_SUCCESS) {
+		afc_unlock(client);
+		return AFC_E_NOT_ENOUGH_DATA;
+	}
+	/* Receive response */
+	ret = afc_receive_data(client, NULL, &bytes);
+
+	afc_unlock(client);
+
+	return ret;
+}
+
+LIBIMOBILEDEVICE_API afc_error_t afc_remove_path_and_contents(afc_client_t client, const char *path)
 {
-	char *send = (char *) malloc(sizeof(char) * (strlen(path) + 1 + 8));
 	uint32_t bytes = 0;
-	uint64_t mtime_loc = htole64(mtime);
 	afc_error_t ret = AFC_E_UNKNOWN_ERROR;
 
 	if (!client || !path || !client->afc_packet || !client->parent)
@@ -1194,11 +1088,16 @@ afc_error_t afc_set_file_time(afc_client_t client, const char *path, uint64_t mt
 
 	afc_lock(client);
 
+	uint32_t data_len = strlen(path) + 1;
+	if (_afc_check_packet_buffer(client, data_len) < 0) {
+		afc_unlock(client);
+		debug_info("Failed to realloc packet buffer");
+		return AFC_E_NO_MEM;
+	}
+
 	/* Send command */
-	memcpy(send, &mtime_loc, 8);
-	memcpy(send + 8, path, strlen(path) + 1);
-	ret = afc_dispatch_packet(client, AFC_OP_SET_FILE_TIME, send, 8 + strlen(path) + 1, NULL, 0, &bytes);
-	free(send);
+	memcpy(AFC_PACKET_DATA_PTR, path, data_len);
+	ret = afc_dispatch_packet(client, AFC_OP_REMOVE_PATH_AND_CONTENTS, data_len, NULL, 0, &bytes);
 	if (ret != AFC_E_SUCCESS) {
 		afc_unlock(client);
 		return AFC_E_NOT_ENOUGH_DATA;
@@ -1211,7 +1110,7 @@ afc_error_t afc_set_file_time(afc_client_t client, const char *path, uint64_t mt
 	return ret;
 }
 
-afc_error_t afc_dictionary_free(char **dictionary)
+LIBIMOBILEDEVICE_API afc_error_t afc_dictionary_free(char **dictionary)
 {
 	int i = 0;
 
