@@ -1,30 +1,32 @@
- /* 
+/*
  * mobilebackup.h
  * Definitions for the mobilebackup service
- * 
+ *
+ * Copyright (c) 2010-2019 Nikias Bassen, All Rights Reserved.
  * Copyright (c) 2009 Martin Szulecki All Rights Reserved.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2.1 of the License, or (at your option) any later version.
- * 
+ *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA 
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  */
-#ifndef MOBILEBACKUP_H
-#define MOBILEBACKUP_H
+
+#ifndef __MOBILEBACKUP_H
+#define __MOBILEBACKUP_H
 
 #include "libimobiledevice/mobilebackup.h"
 #include "device_link_service.h"
 
-struct mobilebackup_client_int {
+struct mobilebackup_client_private {
 	device_link_service_client_t parent;
 };
 
